
#include <stdint.h>
#include <string.h>
#include <rc.h>
#include <rc_adc.h>
#include <spi.h>
#include <font.h>
#include <icons.h>
#include <draw2d.h>
#include <lcd.h>
#include <misc.h>

#ifdef SUPPORT_LCD

uint8_t Framebuffer[((DISPLAY_W * DISPLAY_H) / 8)];
uint8_t *font[] = {(uint8_t *)&au8FontSystem3x6, (uint8_t *)&au8FontSystem5x8, (uint8_t *)&au8FontSystem7x8};
const uint8_t font_w[] = {3, 5, 7};
const uint8_t font_h[] = {6, 8, 8};
uint8_t font_num = 0;


enum {
	SPIDISP_CLEAR,
	SPIDISP_PIXEL,		// X, Y
	SPIDISP_LINE,		// X1, Y1, X2, Y2
	SPIDISP_RECTANGLE,		// X1, Y1, X2, Y2, FILLED
	SPIDISP_CIRCLE,		// X, Y, R, FILLED
	SPIDISP_CHAR,		// X, Y, CH
	SPIDISP_TEXT,		// X, Y, LEN, STRING
	SPIDISP_FONT_TYPE,		// NUM
	SPIDISP_FOREGROUND_COLOR,	// R, G, B
	SPIDISP_BACKGROUND_COLOR,	// R, G, B
	SPIDISP_MAX,
};

#define SPIDISP_CMDLEN_CLEAR			1 + 0
#define SPIDISP_CMDLEN_PIXEL			1 + 2
#define SPIDISP_CMDLEN_LINE			1 + 4
#define SPIDISP_CMDLEN_RECTANGLE		1 + 5
#define SPIDISP_CMDLEN_CIRCLE		1 + 4
#define SPIDISP_CMDLEN_CHAR			1 + 3
#define SPIDISP_CMDLEN_TEXT			1 + 4
#define SPIDISP_CMDLEN_FONT_TYPE		1 + 1
#define SPIDISP_CMDLEN_FOREGROUND_COLOR	1 + 3
#define SPIDISP_CMDLEN_BACKGROUND_COLOR	1 + 3

const uint8_t SPIDISP_LEN[] = {SPIDISP_CMDLEN_CLEAR, SPIDISP_CMDLEN_PIXEL, SPIDISP_CMDLEN_LINE, SPIDISP_CMDLEN_RECTANGLE, SPIDISP_CMDLEN_CIRCLE, SPIDISP_CMDLEN_CHAR, SPIDISP_CMDLEN_TEXT, SPIDISP_CMDLEN_FONT_TYPE, SPIDISP_CMDLEN_FOREGROUND_COLOR, SPIDISP_CMDLEN_BACKGROUND_COLOR};


void Draw2D_Font_SetNum(uint8_t num) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_FONT_TYPE);
	DISPLAY_Write(num);
	DISPLAY_SEL_HI;

	font_num = num;
}

inline void Draw2D_Clear (void) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_CLEAR);
	DISPLAY_SEL_HI;

	memset(&Framebuffer, 0, (DISPLAY_W * DISPLAY_H) / 8);
}

void Draw2D_Pixel (uint8_t x, uint8_t y, uint8_t data) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_PIXEL);
	DISPLAY_Write(x);
	DISPLAY_Write(y);
	DISPLAY_SEL_HI;

	uint16_t index = ((y / 8) * DISPLAY_W) + x;
	if (x < DISPLAY_W && y < DISPLAY_H) {
		uint8_t offset = y - ((y / 8) * 8);
		switch (data) {
			case LCD_FG:
				Framebuffer[index] |= (1<<offset);
			break;
			case LCD_BG:
				Framebuffer[index] &= ~(1<<offset);
			break;
			case LCD_XOR:
				if (Framebuffer[index] & (1<<offset)) {
					Framebuffer[index] &= ~(1<<offset);
				} else {
					Framebuffer[index] |= (1<<offset);
				}
			break;
		}
	}
}

uint8_t Draw2D_Char (uint8_t x, uint8_t y, uint8_t ch, uint8_t fg, uint8_t bg) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_CHAR);
	DISPLAY_Write(x);
	DISPLAY_Write(y);
	DISPLAY_Write(ch);
	DISPLAY_SEL_HI;

	uint8_t n = 0;
	uint8_t n2 = 0;
	uint8_t ret = 0;
	uint8_t y_offset = 8 - font_h[font_num];
	if (x + font_w[font_num] < DISPLAY_W) {
		if (ch >= 32) {
			ch -= 32;
			for (n = 0; n < font_w[font_num]; n++) {
				for (n2 = 0; n2 < font_h[font_num]; n2++) {
					if (font[font_num][ch * font_w[font_num] + n] & (1<<(n2 + y_offset))) {
						ret = n;
						Draw2D_Pixel(x + n, y + n2, fg);
					} else if (bg != LCD_TRANS) {
						Draw2D_Pixel(x + n, y + n2, bg);
					}
				}
			}
			if (bg != LCD_TRANS) {
				for (n2 = 0; n2 < font_h[font_num]; n2++) {
					Draw2D_Pixel(x + n, y + n2, bg);
				}
			}
		}
	}
	return ret;
}

uint8_t Draw2D_CharWidth (char ch) {
	uint8_t n = 0;
	uint8_t ret = 0;
	for (n = 0; n < font_w[font_num]; n++) {
		if (font[font_num][(ch - 32) * font_w[font_num] + n] != 0) {
			ret = n;
		}
	}
	return ret;
}

uint8_t Draw2D_StringWidth (char *text) {
	uint8_t n = 0;
	uint8_t w = 0;
	uint8_t ret = 0;
	while (*text) {
		w = 0;
		for (n = 0; n < font_w[font_num]; n++) {
			if (font[font_num][((*text) - 32) * font_w[font_num] + n] != 0) {
				w = n;
			}
		}
		*text++;
		ret += w + 2;
	}
	return ret;
}

uint8_t Draw2D_StringCenter (uint8_t y, char *text, uint8_t fg, uint8_t bg) {
	uint8_t x = 0;
	char *text_ptr = text;
	while (*text_ptr) {
		x += Draw2D_CharWidth(*text_ptr++) + 2;
	}
	text_ptr = text;
	x = (DISPLAY_W - x) / 2;
	while (*text_ptr) {
		x += Draw2D_Char(x, y, (*text_ptr++), fg, bg) + 2;
	}
	return x;
}

uint8_t Draw2D_StringRight (uint8_t y, char *text, uint8_t fg, uint8_t bg) {
	uint8_t x = 0;
	char *text_ptr = text;
	while (*text_ptr) {
		x += Draw2D_CharWidth(*text_ptr++) + 2;
	}
	text_ptr = text;
	x = DISPLAY_W - x;
	while (*text_ptr) {
		x += Draw2D_Char(x, y, (*text_ptr++), fg, bg) + 2;
	}
	return x;
}

uint8_t Draw2D_PrintXY (uint8_t x, uint8_t y, char *text, uint8_t fg, uint8_t bg) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_TEXT);
	DISPLAY_Write(x);
	DISPLAY_Write(y);
	DISPLAY_Write(strlen(text));
	while (*text) {
		DISPLAY_Write(*text++);
	}
	DISPLAY_SEL_HI;

	if (x == 255) {
		return Draw2D_StringCenter(y, text, fg, bg);
	} else if (x == 254) {
		return Draw2D_StringRight(y, text, fg, bg);
	}
	while (*text) {
		x += Draw2D_Char(x, y, (*text++), fg, bg) + 2;
	}
	return x;
}

void Draw2D_Icon (uint8_t x, uint8_t y, uint8_t w, uint8_t h, char *icon_data, char color) {
	uint8_t x1 = 0;
	uint8_t y1 = 0;
	for (y1 = 0; y1 < h; y1++) {
		for (x1 = 0; x1 < w; x1++) {
			if (icon_data[(y1 * w) + x1] != 0) {
				Draw2D_Pixel(x + x1, y + y1, color);
			}
		}
	}
}

void Draw2D_Rec (uint16_t xstart, uint16_t ystart, uint16_t width, uint16_t height) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_RECTANGLE);
	DISPLAY_Write(xstart);
	DISPLAY_Write(ystart);
	DISPLAY_Write(xstart + width);
	DISPLAY_Write(ystart + height);
	DISPLAY_Write(0); // not FILLED
	DISPLAY_SEL_HI;

	uint16_t x = 0;
	uint16_t y = 0;
	for (y = ystart; y <= ystart + height; y++) {
		Draw2D_Pixel(xstart, y, LCD_FG);
		Draw2D_Pixel(xstart + width, y, LCD_FG);
	}
	for (x = xstart; x <= xstart + width; x++) {
		Draw2D_Pixel(x, ystart, LCD_FG);
		Draw2D_Pixel(x, ystart + height, LCD_FG);
	}
}

void Draw2D_Box (uint16_t xstart, uint16_t ystart, uint16_t width, uint16_t height) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_RECTANGLE);
	DISPLAY_Write(xstart);
	DISPLAY_Write(ystart);
	DISPLAY_Write(xstart + width);
	DISPLAY_Write(ystart + height);
	DISPLAY_Write(1); // FILLED
	DISPLAY_SEL_HI;

	uint16_t x = 0;
	uint16_t y = 0;
	for (y = ystart; y < ystart + height; y++) {
		for (x = xstart; x < xstart + width; x++) {
			Draw2D_Pixel(x, y, LCD_FG);
		}
	}
}

void Draw2D_BoxInverted (uint16_t xstart, uint16_t ystart, uint16_t width, uint16_t height) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_RECTANGLE);
	DISPLAY_Write(xstart);
	DISPLAY_Write(ystart);
	DISPLAY_Write(xstart + width);
	DISPLAY_Write(ystart + height);
	DISPLAY_Write(1); // FILLED
	DISPLAY_SEL_HI;

	uint16_t x = 0;
	uint16_t y = 0;
	for (y = ystart; y < ystart + height; y++) {
		for (x = xstart; x < xstart + width; x++) {
			Draw2D_Pixel(x, y, LCD_BG);
		}
	}
}

void Draw2D_Bar (uint8_t pos_x, uint8_t pos_y, uint8_t w, uint8_t h, uint8_t percent, char *text) {
	uint8_t x = 0;
	uint8_t y = pos_y;
	char *text_ptr = text;

	Draw2D_Font_SetNum(1);

	Draw2D_BoxInverted(pos_x, pos_y, w, h);
	Draw2D_Box(pos_x, pos_y, percent * w / 100, h);
	Draw2D_Rec(pos_x, pos_y, w, h);
	while (*text_ptr) {
		x += Draw2D_CharWidth(*text_ptr++) + 2;
	}
	text_ptr = text;
	x = (w - x) / 2;
	while (*text_ptr) {
		x += Draw2D_Char(pos_x + x, y + 2, (*text_ptr++), LCD_XOR, LCD_TRANS) + 2;
	}
}

void DISPLAY_Header (char *text) {
	uint8_t pos_x = 0;
	char tmp_str[10];

	Draw2D_Font_SetNum(2);
	Draw2D_StringCenter(0, text, LCD_XOR, LCD_TRANS);

	Draw2D_Font_SetNum(1);
	int2str_new((char *)tmp_str, (SystemTimer_time[0] / 1000) / 60, 1, 0, 1);
	pos_x = Draw2D_PrintXY(pos_x, 0, (char *)tmp_str, LCD_XOR, LCD_TRANS);
	pos_x = Draw2D_PrintXY(pos_x, 0, (char *)":", LCD_XOR, LCD_TRANS);
	int2str_new((char *)tmp_str, (uint16_t)((SystemTimer_time[0] / 1000) - ((SystemTimer_time[0] / 1000) / 60 * 60)), 1, 0, 1);
	pos_x = Draw2D_PrintXY(pos_x, 0, (char *)tmp_str, LCD_XOR, LCD_TRANS);

	uint16_t volt = GetADCChanel(ADC_CHANNELS_MAX - 2) * 1540 / 4096;
	int2str_new((char *)tmp_str, (int16_t)volt, 0, 0, 1);
	pos_x = Draw2D_StringWidth((char *)tmp_str);
	pos_x += Draw2D_StringWidth((char *)"V");
	pos_x = Draw2D_PrintXY(DISPLAY_W - pos_x, 0, (char *)tmp_str, LCD_XOR, LCD_TRANS);
	pos_x = Draw2D_PrintXY(pos_x, 0, (char *)"V", LCD_XOR, LCD_TRANS);

	Draw2D_Font_SetNum(1);
}

void Draw2D_Line (uint8_t px1, uint8_t py1, uint8_t px2, uint8_t py2, uint8_t color) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_LINE);
	DISPLAY_Write(px1);
	DISPLAY_Write(py1);
	DISPLAY_Write(px2);
	DISPLAY_Write(py2);
	DISPLAY_SEL_HI;

	int16_t x, y, x1, x2, y1, y2, m, ofs;
	uint8_t counter;
	x1 = (uint16_t)px1;
	x2 = (uint16_t)px2;
	y1 = (uint16_t)py1;
	y2 = (uint16_t)py2;
	if (x2 < x1) {
		x = x1; x1 = x2; x2 = x;
		y = y1; y1 = y2; y2 = y;
	}
	if (y2 == y1) {
		for (counter = x1;counter <= x2;counter++) {
			Draw2D_Pixel((int8_t)counter, (int8_t)y1, color);
		}
	} else{
		if (x2 == x1) {
			if (y2 < y1) {
				x = x1;
				x1 = x2;
				x2 = x;
				y = y1;
				y1 = y2;
				y2 = y;
			}
			for (counter = y1; counter<= y2; counter++) {
				Draw2D_Pixel((uint8_t)x1, (uint8_t)counter, color);
			}
		} else {
			m = (127 * (y2 - y1)) / (x2 - x1);
			if ((m < 127) && (m > - 127)) {
				m = (500 * (y2 - y1)) / (x2 - x1);
				if (m >= 0) {
					ofs = 249;
				} else {
					ofs = -249;
				}
				for (counter = 0; counter <= (x2 - x1); counter++) {
					y = (counter*m + ofs) / 500;
					Draw2D_Pixel((uint8_t)counter + (uint8_t)x1, (uint8_t)y + y1, color);
				}
			} else {
				if ((m >= 127) || (m <= - 127)) {
					if (y2 < y1) {
						x = x1; x1 = x2; x2 = x;
						y = y1; y1 = y2; y2 = y;
					}
					m = (255*(y2 - y1)) / (x2 - x1);
					for (counter = 0; counter <= (y2-y1); counter++) {
						x = ((counter*255) + 127) / m;
						Draw2D_Pixel((uint8_t)(x+x1), (uint8_t)counter+(uint8_t)y1, color);
					}
				}
			}
		}
	}
}

void Draw2D_Circle (uint8_t x0, uint8_t y0, uint8_t radius, uint8_t color) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_LINE);
	DISPLAY_Write(x0);
	DISPLAY_Write(y0);
	DISPLAY_Write(radius);
	DISPLAY_Write(0); // not FILLED
	DISPLAY_SEL_HI;

	int8_t f = 1 - radius;
	int8_t ddF_x = 0;
	int8_t ddF_y = -2 * radius;
	int8_t x = 0;
	int8_t y = radius;
	Draw2D_Pixel(x0, y0 + radius, color);
	Draw2D_Pixel(x0, y0 - radius, color);
	Draw2D_Pixel(x0 + radius, y0, color);
	Draw2D_Pixel(x0 - radius, y0, color);
	while(x < y) {
		if(f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x + 1;
		Draw2D_Pixel(x0 + x, y0 + y, color);
		Draw2D_Pixel(x0 - x, y0 + y, color);
		Draw2D_Pixel(x0 + x, y0 - y, color);
		Draw2D_Pixel(x0 - x, y0 - y, color);
		Draw2D_Pixel(x0 + y, y0 + x, color);
		Draw2D_Pixel(x0 - y, y0 + x, color);
		Draw2D_Pixel(x0 + y, y0 - x, color);
		Draw2D_Pixel(x0 - y, y0 - x, color);
	}
}

void Draw2D_CircleFilled (uint8_t x0, uint8_t y0, uint8_t radius, uint8_t color) {
	DISPLAY_SEL_LO;
	DISPLAY_Write(SPIDISP_LINE);
	DISPLAY_Write(x0);
	DISPLAY_Write(y0);
	DISPLAY_Write(radius);
	DISPLAY_Write(1); // not FILLED
	DISPLAY_SEL_HI;


	int8_t f = 1 - radius;
	int8_t ddF_x = 0;
	int8_t ddF_y = -2 * radius;
	int8_t x = 0;
	int8_t y = radius;
	Draw2D_Line(x0,y0 - radius,x0,y0 + radius, color);
	Draw2D_Line(x0 - radius, y0, x0 + radius, y0, color);
	while(x < y) {
		if(f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x + 1;
		Draw2D_Line(x0 + x, y0 - y, x0 + x, y0 + y, color);
		Draw2D_Line(x0 + y, y0 - x, x0 + y, y0 + x, color);
		Draw2D_Line(x0 - x, y0 - y, x0 - x, y0 + y, color);
		Draw2D_Line(x0 - y, y0 - x, x0 - y, y0 + x, color);
	}
}

#endif


